@@ -1,4 +1,4 @@
-/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
+/* vim: set sw=8: -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /* 
  * gnome-cups-printer.c - 
  * Copyright (C) 2002, Ximian, Inc.
@@ -27,6 +27,8 @@
 
 #include <cups/cups.h>
 #include <time.h>
+#include <stdlib.h>
+#include <ctype.h>
 
 #include "util.h"
 #include "gnome-cups-request.h"
@@ -274,6 +276,13 @@ attributes_update_cb (guint id,
 #undef MAP_INT
 #undef MAP_STRING
 
+/* JEG 
+ * Disable this for now.  The default SuSE network config does not
+ * produce a resolvable hostname for the localhost and cups has a bug
+ * that will produce an infinite loop when presented with an invalid
+ * hostname.  This will also produce a hang if the remote printer is
+ * unavailable and we do a syncronous lookup. */
+static gboolean go_directly_to_printer_when_possible = FALSE;
 
 static void
 update_attributes (GnomeCupsPrinter *printer)
@@ -281,18 +290,20 @@ update_attributes (GnomeCupsPrinter *printer)
 	ipp_t *request;
 	char *host;
 	static const char *attributes[] = {
-	  "printer-state", "queued-job-count",
-	  "printer-location", "printer-info",
-	  "printer-state-message", "device-uri",
-	  "printer-state-reasons", "printer-info",
-	  "printer-make-and-model", "printer-uri-supported"
+	  "printer-name",
+	  "printer-state",		"queued-job-count",
+	  "printer-location",		"printer-info",
+	  "printer-state-message",	"device-uri",
+	  "printer-state-reasons",	"printer-info",
+	  "printer-make-and-model",	"printer-uri-supported"
 	};
 
 	if (printer->details->attributes_request_id > 0) {
 		return;
 	}
 	
-	if (printer->details->attributes_set) {
+	if (go_directly_to_printer_when_possible &&
+	    printer->details->attributes_set) {
 		request = gnome_cups_request_new_for_printer (IPP_GET_PRINTER_ATTRIBUTES,
 							      printer);
 		gnome_cups_request_add_requested_attributes (request, 
@@ -323,6 +334,185 @@ update_attributes (GnomeCupsPrinter *printer)
 	g_free (host);
 }
 
+/*
+ * cups_get_dest() is copied from the cups libraries, file cups/dest.c,
+ * which is 
+ *   Copyright 1997-2004 by Easy Software Products.
+ */
+static int				/* O - Number of destinations */
+cups_get_dests(const char  *filename,	/* I - File to read from */
+               int         num_dests,	/* I - Number of destinations */
+               cups_dest_t **dests)	/* IO - Destinations */
+{
+  int		i;			/* Looping var */
+  cups_dest_t	*dest;			/* Current destination */
+  FILE		*fp;			/* File pointer */
+  char		line[8192],		/* Line from file */
+		*lineptr,		/* Pointer into line */
+		*name,			/* Name of destination/option */
+		*instance;		/* Instance of destination */
+  const char	*printer;		/* PRINTER or LPDEST */
+
+
+ /*
+  * Check environment variables...
+  */
+
+  if ((printer = getenv("LPDEST")) == NULL)
+    if ((printer = getenv("PRINTER")) != NULL)
+      if (strcmp(printer, "lp") == 0)
+        printer = NULL;
+
+ /*
+  * Try to open the file...
+  */
+
+  if ((fp = fopen(filename, "r")) == NULL)
+    return (num_dests);
+
+ /*
+  * Read each printer; each line looks like:
+  *
+  *    Dest name[/instance] options
+  *    Default name[/instance] options
+  */
+
+  while (fgets(line, sizeof(line), fp) != NULL)
+  {
+   /*
+    * See what type of line it is...
+    */
+
+    if (strncasecmp(line, "dest", 4) == 0 && isspace(line[4] & 255))
+      lineptr = line + 4;
+    else if (strncasecmp(line, "default", 7) == 0 && isspace(line[7] & 255))
+      lineptr = line + 7;
+    else
+      continue;
+
+   /*
+    * Skip leading whitespace...
+    */
+
+    while (isspace(*lineptr & 255))
+      lineptr ++;
+
+    if (!*lineptr)
+      continue;
+
+    name = lineptr;
+
+   /*
+    * Search for an instance...
+    */
+
+    while (!isspace(*lineptr & 255) && *lineptr && *lineptr != '/')
+      lineptr ++;
+
+    if (!*lineptr)
+      continue;
+
+    if (*lineptr == '/')
+    {
+     /*
+      * Found an instance...
+      */
+
+      *lineptr++ = '\0';
+      instance = lineptr;
+
+     /*
+      * Search for an instance...
+      */
+
+      while (!isspace(*lineptr & 255) && *lineptr)
+	lineptr ++;
+    }
+    else
+      instance = NULL;
+
+    *lineptr++ = '\0';
+
+   /*
+    * See if the primary instance of the destination exists; if not,
+    * ignore this entry and move on...
+    */
+#if 0
+    /* Don't do this here, since we call this with an empty dests array */
+    if (cupsGetDest(name, NULL, num_dests, *dests) == NULL)
+      continue;
+#endif
+
+   /*
+    * Add the destination...
+    */
+
+    num_dests = cupsAddDest(name, instance, num_dests, dests);
+
+    if ((dest = cupsGetDest(name, instance, num_dests, *dests)) == NULL)
+    {
+     /*
+      * Out of memory!
+      */
+
+      fclose(fp);
+      return (num_dests);
+    }
+
+   /*
+    * Add options until we hit the end of the line...
+    */
+
+    dest->num_options = cupsParseOptions(lineptr, dest->num_options,
+                                         &(dest->options));
+
+   /*
+    * Set this as default if needed...
+    */
+
+    if (strncasecmp(line, "default", 7) == 0 && printer == NULL)
+    {
+      for (i = 0; i < num_dests; i ++)
+        (*dests)[i].is_default = 0;
+
+      dest->is_default = 1;
+    }
+  }
+
+ /*
+  * Close the file and return...
+  */
+
+  fclose(fp);      
+
+  return (num_dests);
+}
+
+static int
+parse_lpoptions (cups_dest_t **dests)
+{
+	gchar *filename;
+	const gchar *home;
+	int num_dests;
+
+	num_dests = 0;
+	*dests = NULL;
+
+	home = g_getenv("CUPS_SERVERROOT");
+	if (!home)
+		home = "/etc/cups";
+	filename = g_build_filename (home, "lpoptions", NULL);
+	num_dests = cups_get_dests (filename, num_dests, dests);
+					  
+	g_free (filename);
+	
+	filename = g_build_filename (g_get_home_dir (), ".lpoptions", NULL);
+	num_dests = cups_get_dests (filename, num_dests, dests);
+	g_free (filename);
+
+	return num_dests;
+}
+
 static char *
 get_default (void)
 {
@@ -333,11 +523,11 @@ get_default (void)
 	ipp_attribute_t *attr;
 	GError *error = NULL;
 	int num_dests;
-	cups_dest_t *dests;
+	cups_dest_t *dests = NULL;
 	cups_dest_t *default_dest;
 
 	/* First look in the environment */
-	
+
 	env = g_getenv ("LPDEST");
 	if (env) {
 		return g_strdup (env);
@@ -349,7 +539,7 @@ get_default (void)
 	}
 	
 	/* Then look in .lpoptions */	
-	num_dests = cupsGetDests (&dests);
+	num_dests = parse_lpoptions (&dests);
 	default_dest = cupsGetDest (NULL, NULL, num_dests, dests);
 	if (default_dest) {
 		name = g_strdup (default_dest->name);
@@ -580,7 +770,7 @@ set_timeout (void)
 	} else if (!should_timeout && update_timeout_id) {
 		g_source_remove (update_timeout_id);
 		update_timeout_id = 0;
-	} 
+	}
 }
 
 guint
@@ -685,32 +875,6 @@ gnome_cups_printer_get_existing (const char *printer_name)
 	return NULL;
 }
 
-static gboolean 
-check_exists (const char *printer)
-{
-	GError *error = NULL;
-	ipp_t *request;
-	ipp_t *response;
-	char *printer_uri;
-
-	request = gnome_cups_request_new (IPP_GET_PRINTER_ATTRIBUTES);
-
-	printer_uri = g_strdup_printf ("ipp://localhost/printers/%s", printer);
-	ippAddString (request, IPP_TAG_OPERATION, IPP_TAG_URI,
-		      "printer-uri", NULL, printer_uri);
-	response = gnome_cups_request_execute (request, NULL, "/", &error);
-	g_free (printer_uri);
-
-	ippDelete (response);
-
-	if (error) {
-		g_error_free (error);
-		return FALSE;
-	} else {
-		return TRUE;
-	}
-}
-
 GnomeCupsPrinter *
 gnome_cups_printer_get (const char *printer_name)
 {
@@ -726,7 +890,7 @@ gnome_cups_printer_get (const char *printer_name)
 		printer_name = default_printer;
 	}
 
-	if (!check_exists (printer_name)) {
+	if (!name_in_list (printer_name, printer_names)) {
 		return NULL;
 	}
 	
@@ -761,6 +925,36 @@ gnome_cups_printer_get_name (GnomeCupsPrinter *printer)
 	return printer->details->printer_name;
 }
 
+#if 0
+CUPS does not support this
+void
+gnome_cups_printer_set_name (GnomeCupsPrinter *printer,
+			     const char *name,
+			     GError **error)
+{
+	ipp_t *request;
+	ipp_t *response;
+
+	g_return_if_fail (GNOME_CUPS_IS_PRINTER (printer));
+	g_return_if_fail (name != NULL);
+
+	if (!strcmp (name, printer->details->printer_name))
+		return;
+
+	request = gnome_cups_request_new_for_printer (IPP_SET_PRINTER_ATTRIBUTES, printer);
+	ippAddString (request,	IPP_TAG_PRINTER, IPP_TAG_NAME,
+		"printer-name", NULL, gnome_cups_strdup (name));
+	response = gnome_cups_request_execute (request, NULL, "/admin/", error);
+	ippDelete (response);
+
+#if 0
+	g_free (printer->details->printer_name);
+	printer->details->printer_name = g_strdup (name);
+#endif
+	update_attributes (printer);
+}
+#endif
+
 gboolean
 gnome_cups_printer_is_gone (GnomeCupsPrinter *printer)
 {
@@ -857,33 +1051,186 @@ gnome_cups_printer_get_full_state (GnomeCupsPrinter *printer)
 	return printer->details->full_state;
 }
 
+static char *
+get_ppd_uri_path (GnomeCupsPrinter *printer)
+{
+	return g_strdup_printf ("/printers/%s.ppd", printer->details->printer_name);
+}
+
+static int
+get_tmp_ppd_file (GnomeCupsPrinter *printer,
+		  char **filename,
+		  GError **error)
+{
+	int fd;
+	char *tmpname;
+	char *tmp_filename;
+
+	tmpname = gnome_cups_util_escape_uri_string (printer->details->printer_name, GNOME_CUPS_UNSAFE_ALL);
+	tmp_filename = g_strdup_printf ("%s-printer-ppd-%s-XXXXXX",
+					g_get_user_name (),
+					tmpname);
+	g_free (tmpname);
+	fd = g_file_open_tmp (tmp_filename, filename, error);
+	g_free (tmp_filename);
+	return fd;
+}
+
+static GHashTable *get_ppd_options (GnomeCupsPrinter *printer, ppd_file_t *ppd);
+
 ppd_file_t *
 gnome_cups_printer_get_ppd (GnomeCupsPrinter *printer)
 {
-	const char *filename;
+	GError *error = NULL;
+	int fd;
+	char *host;
+	char *ppdpath;
+	char *filename;
 	ppd_file_t *ppd;
 
 	g_return_val_if_fail (GNOME_CUPS_IS_PRINTER (printer), NULL);
 
-	filename = cupsGetPPD (printer->details->printer_name);
+	fd = get_tmp_ppd_file (printer, &filename, &error);
+	if (error != NULL) {
+		g_warning ("Couldn't create temporary file: %s",
+			   error->message);
+		g_error_free (error);
+		return NULL;
+	}
+
+	host = _gnome_cups_printer_get_host (printer);
+	ppdpath = get_ppd_uri_path (printer);
+
+	gnome_cups_request_file (host, ppdpath, fd, &error);
+	if (error != NULL) {
+		g_warning ("Couldn't retrieve PPD for %s: %s",
+			   printer->details->printer_name,
+			   error->message);
+		g_error_free (error);
+		return NULL;
+	}
+
+	close (fd);
 	ppd = ppdOpenFile (filename);
+	g_free (filename);
 
 	/* This is loaded in to memory now, so we can free it. */
 	unlink (filename);
 
+	if (!printer->details->ppd_options)
+		printer->details->ppd_options = get_ppd_options (printer, ppd);
+
 	return ppd;
 }
 
-const char *
-gnome_cups_printer_get_uri (GnomeCupsPrinter *printer)
+typedef struct
 {
+	GnomeCupsPrinter *printer;
+	GnomeCupsPrinterGetPPDCallback cb;	
+	gpointer user_data;
+	GDestroyNotify destroy_notify;
+	char *filename;
+	int fd;
+} GnomeCupsPrinterGetPPDWrapData;
+
+static void
+free_get_ppd_wrapper_data (gpointer p)
+{
+	GnomeCupsPrinterGetPPDWrapData *data = p;
+	g_object_unref (data->printer);
+	unlink (data->filename);
+	g_free (data->filename);
+	if (data->destroy_notify)
+		data->destroy_notify (data->user_data);
+	close (data->fd);
+	g_free (data);
+}
+
+static void
+wrap_ppd_request_cb (guint id,
+		     const char *path,
+		     ipp_t *response,
+		     GError **error,
+		     gpointer cb_data)
+{
+	GnomeCupsPrinterGetPPDWrapData *data = cb_data;
+	ppd_file_t *ppd;
+
+	if (error && *error != NULL) {
+		data->cb (id, NULL, error, data->user_data);
+		g_clear_error (error);
+	} else {
+		ppd = ppdOpenFile (data->filename);
+		if (!data->printer->details->ppd_options)
+			data->printer->details->ppd_options = 
+				get_ppd_options (data->printer, ppd);
+
+		data->cb (id, ppd, NULL, data->user_data);
+	}
+}
+
+guint
+gnome_cups_printer_get_ppd_async (GnomeCupsPrinter *printer,
+				  GnomeCupsPrinterGetPPDCallback cb,
+				  gpointer user_data,
+				  GDestroyNotify destroy_notify)
+{
+	GError *error = NULL;
+	guint opid;
+	int fd;
+	char *host;
+	char *ppdpath;
+	char *filename;
+	GnomeCupsPrinterGetPPDWrapData *data;
+
+	g_return_val_if_fail (GNOME_CUPS_IS_PRINTER (printer), 0);
+
+	fd = get_tmp_ppd_file (printer, &filename, &error);
+	if (error != NULL) {
+		g_warning ("Couldn't create temporary file: %s",
+			   error->message);
+		g_error_free (error);
+		return 0;
+	}
+
+	host = _gnome_cups_printer_get_host (printer);
+	ppdpath = get_ppd_uri_path (printer);
+
+	data = g_new0 (GnomeCupsPrinterGetPPDWrapData, 1);
+	data->printer = g_object_ref (printer);
+	data->filename = filename;
+	data->fd = fd;
+	data->cb = cb;
+	data->user_data = user_data;
+	data->destroy_notify = destroy_notify;
+
+	opid = gnome_cups_request_file_async (host, ppdpath, fd,
+					      wrap_ppd_request_cb,
+					      data,
+					      free_get_ppd_wrapper_data);
+	g_free (ppdpath);
+	return opid;
+}
+
+const char *
+gnome_cups_printer_get_device_uri (GnomeCupsPrinter const *printer)
+{
+	g_return_val_if_fail (GNOME_CUPS_IS_PRINTER (printer), "");
+	g_return_val_if_fail (printer->details->device_uri, "");
+	return printer->details->device_uri;
+}
+const char *
+gnome_cups_printer_get_uri (GnomeCupsPrinter const *printer)
+{
+	g_return_val_if_fail (GNOME_CUPS_IS_PRINTER (printer), "");
+	g_return_val_if_fail (printer->details->printer_uri, "");
 	return printer->details->printer_uri;
 }
 
 const char *       
 gnome_cups_printer_get_description (GnomeCupsPrinter *printer)
 {
-	g_return_val_if_fail (GNOME_CUPS_IS_PRINTER (printer), NULL);
+	g_return_val_if_fail (GNOME_CUPS_IS_PRINTER (printer), "");
 	g_return_val_if_fail (printer->details->description, "");
 
 	return printer->details->description;
@@ -905,7 +1252,7 @@ gnome_cups_printer_set_description (GnomeCupsPrinter *printer,
 		return;
 	}
 
-	request = gnome_cups_request_new_for_printer (CUPS_ADD_PRINTER,
+	request = gnome_cups_request_new_for_printer (IPP_SET_PRINTER_ATTRIBUTES,
 						      printer);
 	ippAddString (request, IPP_TAG_PRINTER, IPP_TAG_TEXT,
 		      "printer-info", NULL, description);
@@ -937,10 +1284,10 @@ gnome_cups_printer_set_location (GnomeCupsPrinter *printer,
 		return;
 	}
 
-	request = gnome_cups_request_new_for_printer (CUPS_ADD_PRINTER,
-						      printer);
+	request = gnome_cups_request_new_for_printer (
+		IPP_SET_PRINTER_ATTRIBUTES, printer);
 	ippAddString (request, IPP_TAG_PRINTER, IPP_TAG_TEXT,
-		      "printer-location", NULL, location);
+		"printer-location", NULL, location);
 	response = gnome_cups_request_execute (request, NULL, "/admin/", error);
 	ippDelete (response);
 	update_attributes (printer);
@@ -955,8 +1302,7 @@ gnome_cups_printer_pause (GnomeCupsPrinter *printer,
 
 	g_return_if_fail (GNOME_CUPS_IS_PRINTER (printer));
 	
-	request = gnome_cups_request_new_for_printer (IPP_PAUSE_PRINTER,
-						      printer);
+	request = gnome_cups_request_new_for_printer (IPP_PAUSE_PRINTER, printer);
 	response = gnome_cups_request_execute (request, NULL, "/admin/", error);
 	ippDelete (response);
 	update_attributes (printer);
@@ -971,8 +1317,7 @@ gnome_cups_printer_resume (GnomeCupsPrinter *printer,
 
 	g_return_if_fail (GNOME_CUPS_IS_PRINTER (printer));
 	
-	request = gnome_cups_request_new_for_printer (IPP_RESUME_PRINTER,
-						      printer);
+	request = gnome_cups_request_new_for_printer (IPP_RESUME_PRINTER, printer);
 	response = gnome_cups_request_execute (request, NULL, "/admin/", error);
 	ippDelete (response);
 	update_attributes (printer);	
@@ -1082,9 +1427,9 @@ get_option (ppd_option_t *option)
 }
 
 static GHashTable *
-get_ppd_options (GnomeCupsPrinter *printer)
+get_ppd_options (GnomeCupsPrinter *printer,
+		 ppd_file_t       *ppd)
 {
-	ppd_file_t *ppd;
 	int i;
 	GHashTable *options;
 	
@@ -1092,7 +1437,6 @@ get_ppd_options (GnomeCupsPrinter *printer)
 					 NULL,
 					 (GDestroyNotify)gnome_cups_printer_option_free);
 	
-	ppd = gnome_cups_printer_get_ppd (printer);
 	if (!ppd)
 		return options;
 	
@@ -1107,7 +1451,6 @@ get_ppd_options (GnomeCupsPrinter *printer)
 			g_hash_table_insert (options, option->id, option);
 		}
 	}
-	ppdClose (ppd);
 
 	return options;
 }
@@ -1115,40 +1458,43 @@ get_ppd_options (GnomeCupsPrinter *printer)
 static GHashTable *
 get_dest_options (GnomeCupsPrinter *printer)
 {
-	GHashTable *dest_options;
-	int num_dests;
-	cups_dest_t *dests;
+	int num_dests, i;
 	cups_dest_t *dest;
-	
+	cups_dest_t *dests = NULL;
+	GHashTable *dest_options;
+
+	/* just parse lpoptions */
+	num_dests = parse_lpoptions (&dests);
+
+	dest = cupsGetDest (printer->details->printer_name, NULL, num_dests, dests);	
+
 	dest_options = g_hash_table_new_full (g_str_hash, g_str_equal,
 					      g_free, g_free);
-
-	num_dests = cupsGetDests (&dests);
-
-	dest = cupsGetDest (printer->details->printer_name, NULL, 
-			    num_dests, dests);	
-
+	
 	if (dest) {
-		int i;
 		for (i = 0; i < dest->num_options; i++) {
 			g_hash_table_insert (dest_options, 
 					     g_strdup (dest->options[i].name),
 					     g_strdup (dest->options[i].value));
 		}
 	}
+	printer->details->options_invalid = FALSE;
 
 	cupsFreeDests (num_dests, dests);
 
-	printer->details->options_invalid = FALSE;
-
 	return dest_options;
 }
 
 static void
 update_options (GnomeCupsPrinter *printer)
-{
+{ 
 	if (!printer->details->ppd_options) {
-		printer->details->ppd_options = get_ppd_options (printer);
+		ppd_file_t *ppd;
+		
+		ppd = gnome_cups_printer_get_ppd (printer);
+		
+		if (ppd)
+			ppdClose (ppd);
 	}
 	
 	if (printer->details->options_invalid) {
@@ -1225,6 +1571,10 @@ gnome_cups_printer_get_options (GnomeCupsPrinter *printer)
 	return ret;
 }
 
+extern int	cups_get_dests(const char *filename, int num_dests,
+		               cups_dest_t **dests);
+
+
 GnomeCupsPrinterOption *
 gnome_cups_printer_get_option (GnomeCupsPrinter *printer,
 			       const char *id)
@@ -1592,7 +1942,10 @@ gchar *
 _gnome_cups_printer_get_host (GnomeCupsPrinter *printer)
 {
 	gchar *host = NULL;
-	if (printer->details->printer_uri) {
+
+#warning this is broken for smb://user:pass@host/printer urls
+	if (go_directly_to_printer_when_possible &&
+	    printer->details->printer_uri) {
 		gchar *x, *y;
 
 		x = strstr (printer->details->printer_uri, "://");
